package be.iminds.ilabt.jfed.git;

import be.iminds.ilabt.jfed.util.KeyUtil;
import com.google.common.base.Charsets;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import org.eclipse.jgit.api.*;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.lib.Ref;
import org.eclipse.jgit.lib.SubmoduleConfig;
import org.eclipse.jgit.transport.*;
import org.eclipse.jgit.util.FS;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

/**
 * GitFetcher: downloads git repo's locally
 */
public class GitFetcher {
    private static final Logger LOG = LoggerFactory.getLogger(GitFetcher.class);

    @Nullable
    private final GitAuthPreferences fallbackGitAuthPreferences;

    /**
     * @param fallbackGitAuthPreferences GitAuthPreferences to use as fallback, when no gitAuth is specified in the command itself
     */
    public GitFetcher(@Nullable GitAuthPreferences fallbackGitAuthPreferences) {
        this.fallbackGitAuthPreferences = fallbackGitAuthPreferences;
    }

    /**
     * @param gitFetcherCommand
     * @return the dir in which the files from git are downloaded. If subDir is requested, that local dir corresponding to that subDir is returned.
     */
    public File fetch(GitFetcherCommand gitFetcherCommand) {
        final String subDir = gitFetcherCommand.getRepoSubDir();
        final String gitUrl = gitFetcherCommand.getGitUrl();
        final String branch = gitFetcherCommand.getBranch() != null ? gitFetcherCommand.getBranch() : "master";

        File resDir;
        File rootGitDir = gitFetcherCommand.getTargetDir();
        assert rootGitDir != null;
        if (subDir != null && !subDir.trim().equals("/")) {
            resDir = new File(rootGitDir, subDir);
        } else {
            resDir = rootGitDir;
        }

        //create credentialsprovider
        CredentialsProvider gitCredsProvider = null;
        TransportConfigCallback transportConfigCallback = null;


        GitAuthPreferences.GitAuth gitAuthForRepo = gitFetcherCommand.getGitAuth();
        if (!GitAuthPreferences.urlMatchesAuth(gitAuthForRepo, gitUrl)) {
            //make sure the auth we have matches the URL  (don't set SSH auth for HTTPS etc)
            gitAuthForRepo = null;
        }
        if(gitAuthForRepo == null && fallbackGitAuthPreferences != null){
            gitAuthForRepo = fallbackGitAuthPreferences.getGitAuth(gitUrl, GitAuthPreferences.GitAuthMethod.SSH, GitAuthPreferences.GitAuthMethod.USER_PASS);
        }
        if (!GitAuthPreferences.urlMatchesAuth(gitAuthForRepo, gitUrl)) {
            //make sure the auth we have matches the URL  (don't set SSH auth for HTTPS etc)
            gitAuthForRepo = null;
        }

        if (gitAuthForRepo != null) {

            if (gitAuthForRepo.getMethod() == GitAuthPreferences.GitAuthMethod.SSH) {
                GitAuthPreferences.GitAuthSsh gitAuthSsh = (GitAuthPreferences.GitAuthSsh) gitAuthForRepo;

                SshSessionFactory sshSessionFactory = new JschConfigSessionFactory() {
                    @Override
                    protected JSch createDefaultJSch(FS fs) throws JSchException {
                        JSch jSch = super.createDefaultJSch(fs);
                        byte[] privkey = gitAuthSsh.getPrivateKey().getBytes(Charsets.UTF_8);
                        byte[] pubkey;
                        try {
                            pubkey = new String(
                                    KeyUtil.privateKeyToAnyPem(KeyUtil.pemToRsaPrivateKey(gitAuthSsh.getPrivateKey(), null))
                            ).getBytes(Charsets.UTF_8);
                        } catch (KeyUtil.PEMDecodingException e) {
                            throw new RuntimeException("Error getting public key from private key", e);
                        }
                        byte[] pass = new byte[0];
                        jSch.addIdentity("jFed user private SSH key", privkey, pubkey, pass);
                        return jSch;
                    }

                    @Override
                    protected void configure(OpenSshConfig.Host host, Session session) {
                        //not a safe solution for host key problem
                        session.setConfig("StrictHostKeyChecking", "no");
                    }
                };

                transportConfigCallback = new TransportConfigCallback() {
                    @Override
                    public void configure(Transport transport) {
                        SshTransport sshTransport = (SshTransport) transport;
                        sshTransport.setSshSessionFactory(sshSessionFactory);
                    }
                };
            }

            if (gitAuthForRepo.getMethod() == GitAuthPreferences.GitAuthMethod.USER_PASS) {
                assert gitAuthForRepo instanceof GitAuthPreferences.GitAuthUserPass;
                GitAuthPreferences.GitAuthUserPass gitAuthUserPass = (GitAuthPreferences.GitAuthUserPass) gitAuthForRepo;
                gitCredsProvider = new UsernamePasswordCredentialsProvider(gitAuthUserPass.getUsername(), gitAuthUserPass.getPassword());
            }
        }

        try {
            Git git = Git.init().setDirectory(rootGitDir).call();


            //add remote
            RemoteAddCommand remoteAddCommand = git.remoteAdd();
            remoteAddCommand.setName("origin");
            remoteAddCommand.setUri(new URIish(gitUrl));
            remoteAddCommand.call();

            //fetch remote
            FetchCommand fetchCommand = git.fetch();
            fetchCommand.setRemote("origin");

            if (gitCredsProvider != null)
                fetchCommand.setCredentialsProvider(gitCredsProvider);
            if (transportConfigCallback != null)
                fetchCommand.setTransportConfigCallback(transportConfigCallback);

            //only checkout branch we need
//            RefSpec refSpec = new RefSpec();
//            refSpec.setSource("refs/heads/" + branch);
//            fetchCommand.setRefSpecs(refSpec);
            fetchCommand.setRecurseSubmodules(SubmoduleConfig.FetchRecurseSubmodulesMode.YES);

            FetchResult fetchResult = fetchCommand.call();

            List<Ref> refs = git.branchList().setListMode(ListBranchCommand.ListMode.ALL).call();

            refs.forEach(ref -> LOG.debug("Got ref {}", ref));

            CheckoutCommand checkoutCommand = git.checkout();
            checkoutCommand.setName("refs/remotes/origin/" + branch);

            //this doesn't appear to work. However, everything works without this (it just downloads the entire repo, but uploads only the subdir)
//            //set sparse checkout
//            if (subDir != null)
//                checkoutCommand.addPath(assureStartsAndEndWithSlash(subDir));

            checkoutCommand.call();

        } catch (GitAPIException e) {
            LOG.error("Error fetching Git repo: " + e.getMessage(), e);
            throw new RuntimeException("Exception fetching git repo:" + e.getMessage(), e);
        } catch (URISyntaxException e) {
            LOG.error("Invalid git repo: " + e.getMessage(), e);
            throw new RuntimeException("Invalid git repo:" + e.getMessage(), e);
        }


        return resDir;
    }

    @Nonnull
    private static String assureStartsAndEndWithSlash(@Nonnull String origPath) {
        String res = origPath;
        if (!res.startsWith("/"))
            res = "/"+res;
        if (!res.endsWith("/"))
            res = res+"/";
        return res;
    }
}
